use crate::renderer::components::{Entity, Material, Mesh, Transform};
use crate::vulkan::context::VulkanContext;
use crate::vulkan::pipeline::{Pipeline, triangle::TrianglePipeline};
use crate::vulkan::renderer::VulkanRenderer;
use crate::vulkan::resources::vertex::{MVPData, Vertex2D};
use std::error::Error;
use std::sync::Arc;
use vulkano::VulkanError;
use vulkano::buffer::{Buffer, BufferCreateInfo, BufferUsage};
use vulkano::command_buffer::{
    AutoCommandBufferBuilder, CommandBufferUsage, PrimaryAutoCommandBuffer,
    RenderingAttachmentInfo, RenderingInfo,
};
use vulkano::descriptor_set::{DescriptorSet, WriteDescriptorSet};
use vulkano::memory::allocator::{AllocationCreateInfo, MemoryTypeFilter};
use vulkano::pipeline::{Pipeline as VulkanPipeline, PipelineBindPoint};
use vulkano::render_pass::{AttachmentLoadOp, AttachmentStoreOp};
use vulkano::swapchain::Surface;
use winit::application::ApplicationHandler;
use winit::event::WindowEvent;
use winit::event_loop::ActiveEventLoop;
use winit::window::WindowId;

pub struct App {
    pub vulkan_context: VulkanContext,
    pub renderer: Option<VulkanRenderer>,
    pub entities: Vec<Entity>,
}

impl App {
    pub fn new(vulkan_context: VulkanContext) -> Self {
        Self {
            vulkan_context,
            renderer: None,
            entities: Vec::new(),
        }
    }

    pub fn setup_test_entities(&mut self) -> Result<(), Box<dyn Error>> {
        // Créer un pipeline de test
        let pipeline = TrianglePipeline::new(&self.vulkan_context.device);

        // Créer un buffer de vertex pour un triangle
        let vertices = [
            Vertex2D {
                position: [-0.5, -0.5],
                color: [1.0, 0.0, 0.0], // Rouge
            },
            Vertex2D {
                position: [0.5, -0.5],
                color: [0.0, 1.0, 0.0], // Vert
            },
            Vertex2D {
                position: [0.0, 0.5],
                color: [0.0, 0.0, 1.0], // Bleu
            },
        ];

        let vertex_buffer =
            Vertex2D::create_buffer(vertices.to_vec(), &self.vulkan_context.memory_allocator)
                .unwrap();

        // Créer un buffer uniform pour les matrices MVP
        let mvp_data = MVPData {
            world: [
                [1.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ],
            view: [
                [1.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ],
            projection: [
                [1.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ],
        };

        let uniform_buffer = Buffer::from_data(
            self.vulkan_context.memory_allocator.clone(),
            BufferCreateInfo {
                usage: BufferUsage::UNIFORM_BUFFER,
                ..Default::default()
            },
            AllocationCreateInfo {
                memory_type_filter: MemoryTypeFilter::PREFER_DEVICE
                    | MemoryTypeFilter::HOST_SEQUENTIAL_WRITE,
                ..Default::default()
            },
            mvp_data,
        )
        .unwrap();

        // Créer un descriptor set de test
        let descriptor_set = DescriptorSet::new(
            self.vulkan_context.descriptor_set_allocator.clone(),
            pipeline
                .get_pipeline()
                .layout()
                .set_layouts()
                .get(0)
                .unwrap()
                .clone(),
            [WriteDescriptorSet::buffer(0, uniform_buffer)],
            [],
        )?;

        let material = Material {
            pipeline: pipeline.get_pipeline().clone(),
            descriptor_set,
        };

        // Créer quelques entités de test
        let mut entities = Vec::new();
        for i in 0..3 {
            entities.push(Entity {
                mesh: Mesh {
                    vertex_buffer: vertex_buffer.clone(),
                    vertex_count: 3,
                    instance_count: 1,
                },
                material: material.clone(),
                transform: Transform {
                    position: [i as f32 * 0.5 - 0.5, 0.0, 0.0],
                    rotation: [0.0, 0.0, 0.0],
                    scale: [1.0, 1.0, 1.0],
                },
            });
        }
        self.entities = entities;

        Ok(())
    }

    pub fn render(
        &self,
        command_buffer: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
    ) -> Result<(), Box<dyn Error>> {
        for entity in &self.entities {
            command_buffer
                .bind_pipeline_graphics(entity.material.pipeline.clone())
                .unwrap()
                .bind_descriptor_sets(
                    PipelineBindPoint::Graphics,
                    entity.material.pipeline.layout().clone(),
                    0,
                    entity.material.descriptor_set.clone(),
                )
                .unwrap()
                .bind_vertex_buffers(0, entity.mesh.vertex_buffer.clone())
                .unwrap();
            unsafe {
                command_buffer
                    .draw(entity.mesh.vertex_count, 1, 0, 0)
                    .unwrap();
            }
        }
        Ok(())
    }
}

impl ApplicationHandler for App {
    fn resumed(&mut self, event_loop: &ActiveEventLoop) {
        let window_attributes = winit::window::Window::default_attributes()
            .with_title("Rust ASH Test")
            .with_inner_size(winit::dpi::PhysicalSize::new(
                f64::from(800),
                f64::from(600),
            ));

        let window = Arc::new(event_loop.create_window(window_attributes).unwrap());

        let surface =
            Surface::from_window(self.vulkan_context.instance.clone(), window.clone()).unwrap();

        self.renderer = Some(VulkanRenderer::new(
            window,
            surface,
            self.vulkan_context.device.clone(),
            self.vulkan_context.queue.clone(),
        ));

        self.setup_test_entities().unwrap();
    }

    fn window_event(&mut self, event_loop: &ActiveEventLoop, _id: WindowId, event: WindowEvent) {
        match event {
            WindowEvent::CloseRequested => {
                log::debug!("The close button was pressed; stopping");
                event_loop.exit();
            }
            WindowEvent::Resized(_) => {
                let renderer = self.renderer.as_mut().unwrap();
                renderer.recreate_swapchain = true;
            }
            WindowEvent::RedrawRequested => {
                let (image_index, acquire_future) =
                    match self.renderer.as_mut().unwrap().begin_frame() {
                        Ok(r) => r,
                        Err(VulkanError::OutOfDate) => return,
                        Err(e) => panic!("failed to acquire next image: {e}"),
                    };

                let mut builder = AutoCommandBufferBuilder::primary(
                    self.vulkan_context.command_buffer_allocator.clone(),
                    self.vulkan_context.queue.queue_family_index(),
                    CommandBufferUsage::OneTimeSubmit,
                )
                .unwrap();

                {
                    builder
                        .begin_rendering(RenderingInfo {
                            color_attachments: vec![Some(RenderingAttachmentInfo {
                                load_op: AttachmentLoadOp::Clear,
                                store_op: AttachmentStoreOp::Store,
                                clear_value: Some([0.0, 0.0, 0.0, 1.0].into()),
                                ..RenderingAttachmentInfo::image_view(
                                    self.renderer.as_ref().unwrap().attachment_image_views
                                        [image_index as usize]
                                        .clone(),
                                )
                            })],
                            ..Default::default()
                        })
                        .unwrap()
                        .set_viewport(
                            0,
                            [self.renderer.as_ref().unwrap().viewport.clone()]
                                .into_iter()
                                .collect(),
                        )
                        .unwrap();
                }

                self.render(&mut builder).unwrap();

                {
                    builder.end_rendering().unwrap();
                }

                self.renderer
                    .as_mut()
                    .unwrap()
                    .end_frame(image_index, acquire_future, builder)
                    .unwrap();
            }
            _ => {}
        }
    }

    fn about_to_wait(&mut self, _event_loop: &ActiveEventLoop) {
        let renderer = self.renderer.as_mut().unwrap();
        renderer.window.request_redraw();
    }
}
