use std::sync::Arc;
use vulkano::command_buffer::{AutoCommandBufferBuilder, PrimaryAutoCommandBuffer};
use vulkano::device::Queue;
use vulkano::image::view::ImageView;
use vulkano::pipeline::graphics::viewport::Viewport;
use vulkano::swapchain::{Surface, Swapchain, SwapchainCreateInfo, SwapchainPresentInfo};
use vulkano::sync::{self, GpuFuture};
use vulkano::{Validated, VulkanError};
use winit::window::Window;

pub struct VulkanRenderer {
    pub window: Arc<Window>,
    pub surface: Arc<Surface>,
    pub swapchain: Arc<Swapchain>,
    pub queue: Arc<Queue>,
    pub attachment_image_views: Vec<Arc<ImageView>>,
    pub previous_frame_end: Option<Box<dyn GpuFuture>>,
    pub recreate_swapchain: bool,
    pub viewport: Viewport,
}

impl VulkanRenderer {
    pub fn new(
        window: Arc<Window>,
        surface: Arc<Surface>,
        device: Arc<vulkano::device::Device>,
        queue: Arc<Queue>,
    ) -> Self {
        let window_size = window.inner_size();
        let surface_formats = device
            .physical_device()
            .surface_formats(&surface, Default::default())
            .unwrap();
        let surface_format = surface_formats[0];

        let (swapchain, images) = Swapchain::new(
            device.clone(),
            surface.clone(),
            SwapchainCreateInfo {
                image_extent: window_size.into(),
                image_usage: vulkano::image::ImageUsage::COLOR_ATTACHMENT,
                image_format: surface_format.0,
                ..Default::default()
            },
        )
        .unwrap();

        let attachment_image_views = images
            .into_iter()
            .map(|image| ImageView::new_default(image).unwrap())
            .collect();

        let viewport = Viewport {
            offset: [0.0, 0.0],
            extent: window_size.into(),
            depth_range: 0.0..=1.0,
        };

        Self {
            window,
            surface,
            swapchain,
            queue,
            attachment_image_views,
            previous_frame_end: Some(sync::now(device).boxed()),
            recreate_swapchain: false,
            viewport,
        }
    }

    pub fn begin_frame(&mut self) -> Result<(u32, Box<dyn GpuFuture>), VulkanError> {
        self.previous_frame_end.as_mut().unwrap().cleanup_finished();

        if self.recreate_swapchain {
            self.recreate_swapchain();
            self.recreate_swapchain = false;
        }

        let (image_index, suboptimal, acquire_future) =
            match vulkano::swapchain::acquire_next_image(self.swapchain.clone(), None)
                .map_err(Validated::unwrap)
            {
                Ok(r) => r,
                Err(VulkanError::OutOfDate) => {
                    self.recreate_swapchain = true;
                    return Err(VulkanError::OutOfDate);
                }
                Err(e) => panic!("failed to acquire next image: {e}"),
            };

        if suboptimal {
            self.recreate_swapchain = true;
        }

        Ok((image_index, acquire_future.boxed()))
    }

    pub fn end_frame(
        &mut self,
        image_index: u32,
        acquire_future: Box<dyn GpuFuture>,
        command_buffer: AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
    ) -> Result<(), VulkanError> {
        let command_buffer = command_buffer.build().unwrap();
        let future = self
            .previous_frame_end
            .take()
            .unwrap()
            .join(acquire_future)
            .then_execute(self.queue.clone(), command_buffer)
            .unwrap()
            .then_swapchain_present(
                self.queue.clone(),
                SwapchainPresentInfo::swapchain_image_index(self.swapchain.clone(), image_index),
            )
            .then_signal_fence_and_flush();

        match future.map_err(Validated::unwrap) {
            Ok(future) => {
                self.previous_frame_end = Some(future.boxed());
                Ok(())
            }
            Err(VulkanError::OutOfDate) => {
                self.recreate_swapchain = true;
                self.previous_frame_end = Some(sync::now(self.queue.device().clone()).boxed());
                Ok(())
            }
            Err(e) => {
                println!("failed to flush future: {e}");
                self.previous_frame_end = Some(sync::now(self.queue.device().clone()).boxed());
                Ok(())
            }
        }
    }

    fn recreate_swapchain(&mut self) {
        let image_extent: [u32; 2] = self.window.inner_size().into();
        if image_extent.contains(&0) {
            return;
        }

        let surface_formats = self
            .queue
            .device()
            .physical_device()
            .surface_formats(&self.surface, Default::default())
            .unwrap();
        let surface_format = surface_formats[0];

        let (new_swapchain, new_images) = self
            .swapchain
            .recreate(SwapchainCreateInfo {
                image_extent,
                image_usage: vulkano::image::ImageUsage::COLOR_ATTACHMENT,
                image_format: surface_format.0,
                ..self.swapchain.create_info()
            })
            .expect("failed to recreate swapchain");

        self.swapchain = new_swapchain;
        self.attachment_image_views = new_images
            .into_iter()
            .map(|image| ImageView::new_default(image).unwrap())
            .collect();
        self.viewport.extent = [image_extent[0] as f32, image_extent[1] as f32];
    }
}
